from __future__ import annotations
from typing import cast, Generic, Sequence, Type, TypeVar

from bevy.providers.protocol import ProviderProtocol
import bevy.context.abstract_context as bc


ProviderType = TypeVar("ProviderType", bound=Type[ProviderProtocol])


class ProviderBuilder(Generic[ProviderType]):
    __match_args__ = "providers", "args", "kwargs"

    def __init__(self, provider: ProviderType, *args, **kwargs):
        self.provider = provider
        self.args = args
        self.kwargs = kwargs

    def bind(self, context: bc.AbstractContext) -> ProviderBuilder[ProviderType]:
        return ProviderBuilder(self._bind(context), *self.args, **self.kwargs)

    def _bind(
        self,
        context: bc.AbstractContext
    ) -> ProviderType:
        if self.provider.__bevy_context__ is context:
            return self.provider

        return self._create_bound_provider_type(context)

    def create_and_insert(self, providers: Sequence[ProviderProtocol]) -> Sequence[ProviderProtocol]:
        return self.provider.create_and_insert(providers, *self.args, **self.kwargs)

    def _create_bound_provider_type(self, context: bc.AbstractContext) -> ProviderType:
        return cast(
            ProviderType,
            type(
                self.provider.__name__,
                (self.provider,),
                {
                    "__bevy_context__": context
                }
            )
        )

    @classmethod
    def create(cls, provider: ProviderType | ProviderBuilder, *args, **kwargs) -> ProviderBuilder:
        match provider:
            case ProviderBuilder():
                return provider

        return ProviderBuilder(provider, *args, **kwargs)


from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Type, TypeVar, Sequence

import bevy.providers as p


KeyObject = TypeVar("KeyObject")
ValueObject = TypeVar("ValueObject")
T = TypeVar("T")


class AbstractContext(ABC):
    @abstractmethod
    def add(
        self,
        obj: ValueObject,
        *,
        use_as: KeyObject | None = None,
        propagate: bool = True
    ):
        ...

    @abstractmethod
    def add_provider(
        self,
        provider: Type[p.ProviderProtocol],
        *args,
        __provider__: p.ProviderProtocol | None = None,
        **kwargs
    ):
        ...

    @abstractmethod
    def bind(self, obj: KeyObject, *, propagate: bool = True) -> KeyObject:
        ...

    @abstractmethod
    def branch(self) -> AbstractContext:
        ...

    @abstractmethod
    def create(
        self,
        obj: KeyObject,
        *args,
        add_to_context: bool = False,
        propagate: bool = True,
        **kwargs
    ) -> ValueObject:
        ...

    @abstractmethod
    def get(
        self,
        obj: KeyObject,
        default: ValueObject | T | None = None,
        *, propagate: bool = True
    ) -> ValueObject | T | None:
        ...

    @abstractmethod
    def get_provider_for(
        self,
        obj: KeyObject,
        *,
        propagate: bool = True
    ) -> p.ProviderProtocol[KeyObject, ValueObject] | None:
        ...

    @abstractmethod
    def has(self, obj: KeyObject, *, propagate: bool = True) -> bool:
        ...

    @abstractmethod
    def has_provider_for(self, obj: KeyObject, *, propagate: bool = True) -> bool:
        ...



from __future__ import annotations
from typing import overload, Protocol, Sequence, TypeVar

import bevy.context.abstract_context as bc


KeyObject = TypeVar("KeyObject")
ValueObject = TypeVar("ValueObject")
T = TypeVar("T")


class ProviderProtocol(Protocol[KeyObject, ValueObject]):
    __bevy_context__: bc.AbstractContext

    def add(self, obj: ValueObject, *, use_as: KeyObject | None = None):
        ...

    def bind_to_context(self, obj: KeyObject, contex) -> KeyObject:
        ...

    def create(self, obj: KeyObject, *args, add: bool = False, **kwargs) -> ValueObject:
        ...

    @overload
    def get(self, obj: KeyObject, default: ValueObject) -> ValueObject:
        ...

    @overload
    def get(self, obj: KeyObject, default: None = None) -> ValueObject | None:
        ...

    @overload
    def get(self, obj: KeyObject, default: T) -> ValueObject | T:
        ...

    def get(self, obj: KeyObject, default: ValueObject | T | None = None) -> ValueObject | T | None:
        ...

    def has(self, obj: KeyObject) -> bool:
        ...

    def supports(self, obj: KeyObject) -> bool:
        ...

    @classmethod
    def create_and_insert(
        cls,
        providers: Sequence[ProviderProtocol],
        *args,
        **kwargs
    ) -> Sequence[ProviderProtocol]:
        ...


"""The context uses a few different concepts to provide a comprehensive, yet intuitive, interface for requesting
dependencies be injected into a class instance.

Most fundamentally it uses a repository to store all instances that have been injected by type that they are associated
with. When a class created by the context
"""
from __future__ import annotations
from typing import ParamSpec, Protocol, Type, TypeVar, Sequence

from bevy.context.abstract_context import AbstractContext
from bevy.context.null_context import NullContext
from bevy.providers import InstanceMatchingProvider, TypeMatchingProvider
from bevy.providers.builder import ProviderBuilder
from bevy.providers.protocol import ProviderProtocol
from bevy.sentinel import sentinel


P = ParamSpec("P")
T = TypeVar("T")
KeyObject = TypeVar("KeyObject")
ValueObject = TypeVar("ValueObject")


ProviderConstructor = ProviderBuilder | Type[ProviderProtocol]


NOT_FOUND = sentinel("NOT_FOUND")


class NoSupportingProviderFoundInContext(Exception):
    ...


class Context(AbstractContext):
    def __init__(self, *providers: ProviderConstructor, parent: Context | None = None):
        self._parent = parent or NullContext()
        self._providers, self._provider_constructors = self._build_providers(providers)

    def add(
        self,
        obj: ValueObject,
        *,
        use_as: KeyObject | None = None,
        propagate: bool = True
    ):
        provider = self.get_provider_for(use_as or obj, propagate=propagate)
        if not provider:
            raise NoSupportingProviderFoundInContext(
                f"No provider was found in the context that supports the object being added. {obj=} {use_as=} "
                f"{propagate=}"
            )

        provider.add(obj, use_as=use_as)

    def add_provider(
        self,
        provider: Protocol[ProviderProtocol],
        *args,
        **kwargs
    ):
        builder = ProviderBuilder.create(provider, *args, **kwargs)
        self._providers = builder.bind(self).create_and_insert(self._providers)
        self._provider_constructors.append(builder)

    def bind(self, obj: KeyObject, *, propagate: bool = True) -> KeyObject:
        provider = self.get_provider_for(obj, propagate=propagate)
        if not provider:
            raise NoSupportingProviderFoundInContext(
                f"No provider was found in the context that supports the object. {obj=} {propagate=}"
            )

        return provider.bind_to_context(obj, self)

    def branch(self) -> Context:
        return type(self)(*self._provider_constructors, parent=self)

    def create(
        self,
        obj: KeyObject,
        *args,
        add_to_context: bool = False,
        propagate: bool = True,
        **kwargs
    ) -> ValueObject:
        provider = self.get_provider_for(obj, propagate=propagate)
        return provider.create(obj, *args, add=add_to_context, **kwargs)

    def get(
        self,
        obj: KeyObject,
        default: ValueObject | T | None = None,
        *, propagate: bool = True
    ) -> ValueObject | T | None:
        provider = self.get_provider_for(obj, propagate=False)
        if provider and provider.has(obj):
            return provider.get(obj)

        if propagate:
            return self._parent.get(obj, default)

        return default

    def get_provider_for(
        self,
        obj: KeyObject,
        *,
        propagate: bool = True
    ) -> ProviderProtocol[KeyObject, ValueObject] | None:
        if p := self._find_provider(obj):
            return p

        if propagate and self._parent.has_provider_for(obj):
            return self._parent.get_provider_for(obj)

    def has(self, obj: KeyObject, *, propagate: bool = True) -> bool:
        return self.get(obj, NOT_FOUND, propagate=propagate) is not NOT_FOUND

    def has_provider_for(self, obj: KeyObject, *, propagate: bool = True) -> bool:
        return self.get_provider_for(obj, propagate=propagate) is not None

    def _build_providers(
        self,
        provider_types: Sequence[ProviderConstructor]
    ) -> tuple[Sequence[ProviderProtocol], list[ProviderBuilder]]:
        if not provider_types:
            return self._build_providers((InstanceMatchingProvider, TypeMatchingProvider))

        return self._build_providers_from_provider_types(provider_types)

    def _build_providers_from_provider_types(
        self,
        provider_types: Sequence[ProviderConstructor]
    ) -> tuple[Sequence[ProviderProtocol], list[ProviderBuilder]]:
        builders = []
        providers = ()
        for provider in provider_types:
            builders.append(builder := ProviderBuilder.create(provider))
            providers = builder.bind(self).create_and_insert(providers)

        return providers, builders

    def _find_provider(self, obj: KeyObject) -> ProviderProtocol | None:
        for provider in self._providers:
            if provider.supports(obj):
                return provider



from __future__ import annotations
from typing import cast, Generic, overload, Type, TypeVar

from bevy.inject.annotations import AnnotationGetter
from bevy.inject.context_injector import ContextInjector
from bevy.inject.inject_strategies import (
    InjectionStrategy,
    InjectAllStrategy,
    InjectAllowStrategy,
    InjectDisallowStrategy
)


T = TypeVar("T")


class Detect:
    ALL = InjectAllStrategy()
    ONLY = InjectAllowStrategy
    IGNORE = InjectDisallowStrategy


class Inject:
    ...


class Bevy:
    __bevy_context__ = None
    _detection_strategy = Detect.ONLY()
    bevy = ContextInjector()

    def __init_subclass__(cls, **kwargs):
        dependencies = cls._detection_strategy.get_declared_dependencies(cls)
        cls._detection_strategy.create_injectors(cls, dependencies)

    def __class_getitem__(cls, strategy: InjectionStrategy) -> Type[Bevy]:
        return cast(
            Type[Bevy],
            type(
                cls.__name__,
                (cls,),
                {
                    "_detection_strategy": strategy
                }
            )
        )


class InjectionDescriptor(Generic[T]):
    def __init__(self, on_cls: Type[Bevy], attr_name: str, annotation_getter: AnnotationGetter[Type[T], T]):
        self.on_cls = on_cls
        self.attr_name = attr_name
        self.annotation_getter = annotation_getter

    @overload
    def __get__(self, instance: Bevy, owner) -> T:
        ...

    @overload
    def __get__(self, instance: None, owner) -> InjectionDescriptor:
        ...

    def __get__(self, instance: Bevy | None, owner) -> T | InjectionDescriptor:
        if not instance:
            return self

        type_hint = self.annotation_getter.get()
        return instance.bevy.get(type_hint) or instance.bevy.create(type_hint, add_to_context=True)


from bevy.context.abstract_context import AbstractContext


class NullContext(AbstractContext):
    def add(self, *_, **__) -> None:
        return None

    def add_provider(self, *_, **__) -> None:
        return None

    def bind(self, *_, **__) -> None:
        return

    def branch(self) -> AbstractContext:
        return self

    def create(self, *_, **__) -> None:
        return None

    def get(self, *_, **__) -> None:
        return None

    def get_provider_for(self, *_, **__) -> None:
        return None

    def has(self, *_, **__) -> None:
        return

    def has_provider_for(self, *_, **__) -> None:
        return

    def __bool__(self):
        return False



from typing import Sequence, TypeVar

from bevy.inject import Bevy
from bevy.providers.protocol import ProviderProtocol
from bevy.sentinel import sentinel


KeyObject = TypeVar("KeyObject")
ValueObject = TypeVar("ValueObject")
T = TypeVar("T")


NOT_FOUND = sentinel("NOT_FOUND")


class InstanceMatchingProvider(ProviderProtocol, Bevy):
    def __init__(self, *_, **__):
        super().__init__()
        self._repository = {}

    def add(self, obj: ValueObject, *, use_as: KeyObject | None = None):
        self._repository[use_as or obj] = obj

    def bind_to_context(self, obj: KeyObject, context) -> KeyObject:
        raise Exception("Cannot bind instances to a context")

    def create(self, obj: KeyObject, *args, add: bool = False, **kwargs) -> ValueObject:
        if add:
            self.add(obj)

        return obj

    def get(self, obj: KeyObject, default: ValueObject | T | None = None) -> ValueObject | T | None:
        for key, value in self._repository.items():
            if obj is key:
                return value

        return default

    def has(self, obj: KeyObject) -> bool:
        return self.get(obj, NOT_FOUND) is not NOT_FOUND

    def supports(self, obj: KeyObject) -> bool:
        return not isinstance(obj, type)

    @classmethod
    def create_and_insert(
        cls,
        providers: Sequence[ProviderProtocol],
        *args,
        **kwargs
    ) -> Sequence[ProviderProtocol]:
        return cls(*args, **kwargs), *providers



from typing import Sequence, TypeVar

from bevy.inject import Bevy
from bevy.providers.protocol import ProviderProtocol
from bevy.sentinel import sentinel


KeyObject = TypeVar("KeyObject")
ValueObject = TypeVar("ValueObject")
T = TypeVar("T")


NOT_FOUND = sentinel("NOT_FOUND")


class TypeMatchingProvider(ProviderProtocol, Bevy):
    def __init__(self, *_, **__):
        super().__init__()
        self._repository = {}

    def add(self, obj: ValueObject, *, use_as: KeyObject | None = None):
        key = use_as or type(obj)
        self._repository[key] = obj

    def bind_to_context(self, obj: KeyObject | type, context) -> KeyObject | type:
        return type(obj.__name__, (obj,), {"__bevy_context__": context})

    def create(self, obj: KeyObject, *args, add: bool = False, **kwargs) -> ValueObject:
        value = self.bevy.bind(obj)(*args, **kwargs)
        if add:
            self.add(value)

        return value

    def get(self, obj: KeyObject, default: ValueObject | T | None = None) -> ValueObject | T | None:
        for key, value in self._repository.items():
            if obj is key or (isinstance(obj, type) and (issubclass(obj, key) or issubclass(key, obj))):
                return value

        return default

    def has(self, obj: KeyObject) -> bool:
        return self.get(obj, NOT_FOUND) is not NOT_FOUND

    def supports(self, obj: KeyObject) -> bool:
        return isinstance(obj, type)

    @classmethod
    def create_and_insert(
        cls,
        providers: Sequence[ProviderProtocol],
        *args,
        **kwargs
    ) -> Sequence[ProviderProtocol]:
        return *providers, cls(*args, **kwargs)



from functools import cache
from typing import Type


@cache
def sentinel(name: str, truthy: bool = False) -> Type:
    class SentinelMCS(type):
        def __bool__(self):
            return truthy

        def __eq__(self, other):
            return False

        def __hash__(self):
            return id(self)

        def __repr__(self):
            return f"<{self.__name__}>"

    return SentinelMCS(
        name,
        (object,),
        {}
    )



from __future__ import annotations
from functools import cache
from typing import Any, Generic, get_type_hints, TypeVar

import bevy.inject.inject as i


AnnotationType = TypeVar("AnnotationType", bound=type)
T = TypeVar("T")


class AnnotationGetter(Generic[AnnotationType, T]):
    def __init__(self, owner_cls: type, attr_name: str, annotation: AnnotationType[T], value: i.Inject | None):
        self.annotation = annotation
        self.attr_name = attr_name
        self.owner_cls = owner_cls
        self.value = value

    def get(self) -> AnnotationType[T]:
        return self.annotation

    @classmethod
    def factory(cls, owner_cls: type, attr_name: str, annotation: Any, value: i.Inject | None) -> AnnotationGetter:
        if isinstance(annotation, str):
            return LazyAnnotationGetter(owner_cls, attr_name, annotation, value)

        return AnnotationGetter(owner_cls, attr_name, annotation, value)


class LazyAnnotationGetter(AnnotationGetter):
    def __init__(self, owner_cls: type, attr_name: str, annotation: str, value: i.Inject | None):
        super().__init__(owner_cls, attr_name, annotation, value)

    @cache
    def get(self) -> type:
        type_hints = get_type_hints(self.owner_cls)
        return type_hints[self.attr_name]


from __future__ import annotations
from functools import cache
from typing import overload, Type, TypeVar

import bevy.context.abstract_context as bc
import bevy.context as c


T = TypeVar("T")


class ContextInjector:
    @overload
    def __get__(self, instance: None, owner: Type[T]) -> None:
        ...

    @overload
    def __get__(self, instance: T, owner: Type[T]) -> bc.AbstractContext:
        ...

    @cache
    def __get__(self, instance: T | None, owner: Type[T]) -> bc.AbstractContext | None:
        if not instance:
            return

        if instance.__bevy_context__:
            return instance.__bevy_context__

        return c.Context()



from __future__ import annotations
from abc import ABC, abstractmethod
from inspect import get_annotations
from typing import Type


from bevy.inject.annotations import AnnotationGetter
import bevy.inject.inject as i


def _is_inject(obj) -> bool:
    return obj is i.Inject or isinstance(obj, i.Inject) or (isinstance(obj, type) and issubclass(obj, i.Inject))


def _make_set(item) -> set:
    if isinstance(item, tuple):
        return set(item)

    return {item}


class InjectionStrategy(ABC):
    @abstractmethod
    def get_declared_dependencies(self, t: type) -> dict[str, AnnotationGetter]:
        ...

    def create_injectors(self, on_cls: Type[i.Bevy], dependencies: dict[str, AnnotationGetter]):
        for name, annotation_getter in dependencies.items():
            setattr(on_cls, name, i.InjectionDescriptor(on_cls, name, annotation_getter))


class InjectAllStrategy(InjectionStrategy):
    """ This will scan a class's attribute annotations and create injection descriptors for any that aren't already
    assigned. """
    def get_declared_dependencies(self, t: type) -> dict[str, AnnotationGetter]:
        return {
            name: AnnotationGetter.factory(t, name, annotation, value)
            for name, annotation in get_annotations(t).items()
            if (
                (value := getattr(t, name, None)) is None
                or _is_inject(value)
            )
        }


class InjectAllowStrategy(InjectionStrategy):
    """ This will scan a class's attribute annotations and create injection descriptors for any that aren't already
    assigned and that are in the allowed set. """
    def __init__(self, allow: set[str] | None = None):
        super().__init__()
        self.allow = allow or set()

    def __class_getitem__(cls, allowed: tuple[str]) -> InjectAllowStrategy:
        return InjectAllowStrategy(_make_set(allowed))

    def get_declared_dependencies(self, t: type) -> dict[str, AnnotationGetter]:
        return {
            name: AnnotationGetter.factory(t, name, annotation, value)
            for name, annotation in get_annotations(t).items()
            if (
                (value := getattr(t, name, None)) is None
                and name in self.allow
                or _is_inject(value)
            )
        }


class InjectDisallowStrategy(InjectionStrategy):
    """ This will scan a class's attribute annotations and create injection descriptors for any that aren't already
    assigned and that aren't in the disallowed set. """
    def __init__(self, disallow: set[str]):
        super().__init__()
        self.disallow = disallow

    def __class_getitem__(cls, disallow: tuple[str]) -> InjectDisallowStrategy:
        return InjectDisallowStrategy(_make_set(disallow))

    def get_declared_dependencies(self, t: type) -> dict[str, AnnotationGetter]:
        return {
            name: AnnotationGetter.factory(t, name, annotation, value)
            for name, annotation in get_annotations(t).items()
            if (
                (value := getattr(t, name, None)) is None
                and name not in self.disallow
                or _is_inject(value)
            )
        }